from swimport.model.cpp_object import CPPObject, Variable, Function, Container, Enumeration, Typedef,\
    Source, FileSource, RawSource
from swimport.model.rules import TriggerBehaviourSwimRule, Swimporting, Compileable, object_swimporting,\
    SwimRule, Trigger, Behaviour, NameTrigger, SwimportingsResult, SwimportAggregate



from typing import Dict, Union, Callable, Iterable

from abc import ABC

from textwrap import dedent
from datetime import datetime
from functools import partial

from swimport.model import FileSource
import swimport_data as data
import swimport.swim as swim_module


class Pool(ABC):
    """a class for pools or pools with partial arguments"""

    def __init__(self, func: Callable, name, args, kwargs):
        super().__init__()
        self.__func__ = func
        self.args = args
        self.kwargs = kwargs
        self.name = name
        self._doc = None

    def apply(self, swim, args, kwargs):
        ret = self.__func__(*self.args, *args, swim=swim, **kwargs, **self.kwargs)
        if ret is None:
            ret = True
        return ret

    def __call__(self, *args, **kwargs):
        if not (args or kwargs):
            return self
        if args and isinstance(args[-1], swim_module.Swim):
            *args, swim = args
            return self.apply(swim, args, kwargs)
        elif 'swim' in kwargs:
            swim = kwargs.pop('swim')
            return self.apply(swim, args, kwargs)
        else:
            args = self.args + args
            kwargs = {**self.kwargs, **kwargs}
            return type(self)(self.__func__, self.name, args, kwargs)

    @property
    def func(self):
        return partial(self.__func__, *self.args, **self.kwargs)

    @property
    def __doc__(self):
        return self._doc

    @__doc__.setter
    def __doc__(self, v):
        if v is None:
            self._doc = None
        else:
            self._doc = self.name + ':\n\t' + '\n\t'.join(dedent(v).strip().splitlines(keepends=False))

    def __hash__(self):
        return hash((type(self), self.__func__, self.args, tuple(self.kwargs.items())))

    def __eq__(self, other):
        return type(self) == type(other) \
               and (self.__func__, self.args, self.kwargs, self.kwargs) == \
               (other.__func__, other.args, other.kwargs, other.kwargs)

    def __str__(self):
        return 'pool: ' + self.name


class IdiomaticPool(Pool):
    """
    A pool that checks that it hasn't been used before
    """

    def apply(self, swim, args, kwargs):
        if self in swim.imported:
            return False
        ret = super().apply(swim, args, kwargs)
        swim.imported[self] = self
        return ret

    @property
    def __doc__(self):
        return Pool.__doc__.fget(self)

    @__doc__.setter
    def __doc__(self, v):
        Pool.__doc__.fset(self, v)


class Pools(Iterable[Pool]):
    """
    Class for object containing all the pools.
    """

    def __init__(self):
        self.registry: Dict[str, Pool] = {}

    def __contains__(self, item):
        return item in self.registry

    def __iter__(self):
        yield from self.registry.values()

    def __len__(self):
        return len(self.registry)

    def add(self, pool_cls=Pool, **kwargs):
        def ret(func, name=..., doc=...):
            nonlocal pool_cls
            if name is ...:
                name = func.__name__
            if doc is ...:
                doc = func.__doc__
            if pool_cls is ...:
                pool_cls = type(func)
                func = func.func

            to_register = pool_cls(func, name, (), {})
            to_register.__doc__ = doc

            v = self.registry.setdefault(name, to_register)
            setattr(self, name, to_register)
            if v is not to_register:
                raise Exception('duplicate pool name: ' + name)
            return to_register

        return partial(ret, **kwargs)


class DocumentingPools(Pools):
    """
    A pools object that updates its __doc__ when a pool is added to it.
    """

    def __init__(self, preamble: str = ''):
        super().__init__()
        self.auto_build_doc = True  # set this to false before adding a lot of pools, then set it to true again.
        self.__doc__ = ''
        self.preamble = preamble

    def add(self, *args, **kwargs):
        ret_ = super().add(*args, **kwargs)

        def ret(*args, **kwargs):
            r = ret_(*args, **kwargs)
            if self.auto_build_doc:
                self.rebuild_doc()
            return r

        return ret

    def rebuild_doc(self):
        self.__doc__ = self.preamble + '\n\n'.join(p.__doc__ for p in sorted(self, key=lambda p: p.name))


pools = DocumentingPools('The following pools are available:\n\n')
pools.auto_build_doc = False

syspools = Pools()  # contains pools that are meant to pe added programmatically, not manually


@pools.add(IdiomaticPool)
def disclose(swim):
    """
    add a disclosure the file was generated, including the swimport version and the time
    """
    swim.add_comment_ml(f"""
                    This file was programmatically generated with {data.__name__} v{data.__version__}
                    on {datetime.now():%Y-%m-%d %H:%M}
                    """)
    swim.add_nl()


@syspools.add(IdiomaticPool)
def std_exception(swim):
    swim(include("<exception>", "<system_error>", "<stdexcept>", "<typeinfo>", "<utility>"))
    swim.add_begin("""
    #define SWIM_STD_CATCH(ctype, ptype) catch (const ctype& e){PyErr_SetString(ptype, e.what()); SWIG_fail;}
""")

    swim.add_raw("""
%define SWIM_STD_EXCEPTION(pattern)
%exception{
    try {
        $action;
    }
    catch (PyObject* const & e){
        if (e == nullptr)
        {
            PyErr_SetString(PyExc_Exception, "an null of static type python object was thrown");
            SWIG_fail;
        }
        if (e == Py_Ellipsis)
        {
            //raising ... means we have an exception ready, just fail without setting an error
            SWIG_fail;
        }
        if (PyObject_IsInstance(e, PyExc_BaseException))
        {
            PyObject* type = PyObject_Type(e);
            PyObject* args = PyObject_GetAttrString(e, "args");
            PyErr_SetObject(type, args);
            Py_DECREF(type);
            Py_DECREF(args);
            Py_DECREF(e);
            SWIG_fail;
        }
        if (PyObject_IsSubclass(e, PyExc_BaseException) == 1)
        {
            PyErr_SetNone(e);
            Py_DECREF(args);
            SWIG_fail;
        }
        PyObject* args = Py_BuildValue("(sN)", "a non-exception python object was thrown", e);
        PyErr_SetObject(PyExc_Exception, args);
        Py_DECREF(args);
        SWIG_fail;
    }
    catch(std::pair<PyObject*, std::string> const & p){
        std::string msg = "error while running function $symname: ";
        msg += std::get<1>(p);
        PyErr_SetString(std::get<0>(p), msg.c_str()); SWIG_fail;
    }
    catch (const std::system_error& e){            
        PyObject* args = Py_BuildValue("(is)", e.code(), e.what());
        PyErr_SetObject(PyExc_OSError, args);
        Py_DECREF(args);
        SWIG_fail;
    }
    catch (const char*& e){
        PyErr_SetString(PyExc_Exception, e);
        SWIG_fail;
    }
    catch (const int& e){
    #if !defined(_WIN32) && (defined(__unix__) || defined(__unix) || (defined(__APPLE__) && defined(__MACH__)))
        PyObject* args = Py_BuildValue("i", e);
        PyErr_SetObject(PyExc_OSError, args);
        Py_DECREF(args);
    #else
        PyErr_SetFromWindowsErr(e);
    #endif
        SWIG_fail;
    }
    SWIM_STD_CATCH(std::bad_alloc, PyExc_MemoryError)
    SWIM_STD_CATCH(std::bad_cast, PyExc_TypeError)
    SWIM_STD_CATCH(std::bad_typeid, PyExc_TypeError)
    SWIM_STD_CATCH(std::range_error, PyExc_OverflowError)
    SWIM_STD_CATCH(std::overflow_error, PyExc_OverflowError)
    SWIM_STD_CATCH(std::underflow_error, PyExc_OverflowError)
    SWIM_STD_CATCH(std::logic_error, PyExc_ValueError)
    SWIM_STD_CATCH(std::exception, PyExc_Exception)
    catch(...){
        PyErr_SetString(PyExc_Exception, "an non-exception was thrown");
        SWIG_fail;
    }
    
    if (errno){
    #if !defined(_WIN32) && (defined(__unix__) || defined(__unix) || (defined(__APPLE__) && defined(__MACH__)))
        PyErr_SetFromErrno(PyExc_OSError);
    #else
        PyErr_SetFromWindowsErr(errno);
    #endif
        SWIG_fail;
    }
}
%enddef
""")


@pools.add(IdiomaticPool, name='print')
def print_(swim):
    """
    adds the neccessary lines to print (using cout) in the code.
    """
    swim.add_begin("""
                #include <iostream>
                #include <cstdio>
                #define PY_PRINT(preamble, o) wcout << preamble; PyObject_Print(o, stdout, 0); wcout << endl;
                using namespace std;
                """)
    swim.add_nl()


@syspools.add(IdiomaticPool)
def dll_main(swim):
    swim.add_begin('''
            #include <windows.h>
            #include <iostream>
            #define SWIM_LOAD_HANDLE(name) HINSTANCE SWIM_DLL_HANDLE_##name = LoadLibrary(#name ".dll");if(!SWIM_DLL_HANDLE_##name){PyErr_SetString(PyExc_ImportError, "could not load library " #name);return NULL;}
            #define SWIM_DECL_IMPORTED_FUNC(retType,name,...) typedef retType (*SWIM_FUNC_TYPE_##name)(__VA_ARGS__); SWIM_FUNC_TYPE_##name SWIM_FUNC_HANDLE_##name;
            #define SWIM_IMPORT_FROMDLL(func,handleName) SWIM_FUNC_HANDLE_##func = (SWIM_FUNC_TYPE_##func)GetProcAddress(SWIM_DLL_HANDLE_##handleName, #func); if (!SWIM_FUNC_HANDLE_##func){PyErr_SetString(PyExc_ImportError, "could not load function " #func " from " #handleName ".dll"); return NULL;}
        ''')


@syspools.add(IdiomaticPool)
def dll_load(name, *, swim):
    if name.endswith('.dll'):
        name = name[:-len('.dll')]
    swim.add_init(f'SWIM_LOAD_HANDLE({name})')


@pools.add(Pool)
def dlls(*dll_names, swim):
    """
    if you want to import functions from dlls.
    :param dll_names: an optional list of dll names to load.
    """
    swim(dll_main())
    for name in dll_names:
        swim(dll_load(name))
    swim.add_nl()


@pools.add(Pool)
def module(name, docstring=None, package=None, *, swim, **kwargs):
    """
    to specify the module name (if the swim is initialized with a name, this is called automatically)
    :param name: the name of the output module
    :param docstring: the docstring of the module, if any
    :param package: the package of the module, if any
    :param kwargs: additional options for the module
    """
    options = kwargs
    if docstring:
        options['docstring'] = '"' + docstring + '"'
    if package:
        options['package'] = '"' + package + '"'
    if options:
        opt_string = '(' + ', '.join(k + ' = ' + v for k, v in options.items()) + ')'
    else:
        opt_string = ''
    swim.add_raw('%module' + opt_string + ' ' + name)
    swim.add_nl()


@pools.add(IdiomaticPool)
def include(*sources: Union[FileSource, str], swim):
    """
    include a source with its preprocessor directives
    :param sources: the source objects or paths to cpp header files
    """
    lines = []
    for source in sources:
        if isinstance(source, str):
            if source.startswith('<') or source.startswith('"'):
                lines.append('#include ' + source)
                continue
            source = FileSource(source)
        lines.extend('#define ' + directive for directive in source.directives)
        lines.append('#include "' + source.source_path + '"')
        lines.extend('#undef ' + directive for directive in reversed(source.directives))

    swim.add_begin(lines)
    swim.add_nl()


@syspools.add(IdiomaticPool)
def ns_trap_main(swim):
    swim.add_python_begin("""
                    class _SWIMPORT_NamespaceTrap:
                        # an object that, when overridden, deletes its overrider from the namespace
                        def __init__(self, ns, key):
                            self.ns = ns
                            self.key = key
                        def __del__(self):
                            self.ns.pop(self.key, None)
                    """)


@syspools.add(IdiomaticPool)
def ns_trap_name(name, *, swim):
    swim.add_python(f"""
                try:
                    del {name}
                except NameError:
                    {name} = _SWIMPORT_NamespaceTrap(globals(), '''{name}''')
                """)


@syspools.add(IdiomaticPool)
def ns_trap(*names, swim):
    """
    include a python class to delete whatever overrides it in the namespace
    """
    swim(ns_trap_main())
    for name in names:
        swim(ns_trap_name(name))


@syspools.add(IdiomaticPool)
def typemap_macros(swim):
    swim(include("<exception>", "<utility>", "<string>"))
    swim.add_begin("""
    #define PACKED_UD(ind) (((void**)userdata)[(ind)])
    // note: RAISE macros are supposed to be thrown form py->cpp functions, and THROW from cpp->py
    #define SWIM_RAISE(TYPE,MSG) throw std::pair<PyObject*, std::string>((TYPE), (MSG))
    #define SWIM_RAISE_KEY(MSG, KEY)   { std::string msg = (MSG);\\
                                         auto key_str = PyObject_Str((KEY));\\
                                         msg += PyUnicode_AsUTF8(key_str);\\
                                         Py_DECREF(key_str);\\
                                         SWIM_RAISE(PyExc_KeyError, msg); }
    #define SWIM_RAISE_TYPE(MSG, OBJ)  { std::string msg = (MSG);\\
                                         auto otype = PyObject_Type((OBJ)); \\
                                         auto otype_str = PyObject_Str(otype);\\
                                         msg += PyUnicode_AsUTF8(otype_str);\\
                                         Py_DECREF(otype);\\
                                         Py_DECREF(otype_str);\\
                                         SWIM_RAISE(PyExc_TypeError, msg); }
    #define SWIM_RAISE_UNSPECIFIED_VAR(var) SWIM_RAISE_TYPE("an unspecified exception occurred, input type: ", var);
    #define SWIM_RAISE_UNSPECIFIED SWIM_RAISE_UNSPECIFIED_VAR(input);
    #define SWIM_RAISE_EXISTING throw Py_Ellipsis;
    
    #define SWIM_THROW SWIM_RAISE
    #define SWIM_THROW_TYPE(MSG, OBJ)  { std::string msg = (MSG);\\
                                         msg += typeid((OBJ)).name();\\
                                         SWIM_RAISE(PyExc_TypeError, msg); }
    #define SWIM_THROW_UNSPECIFIED_VAR(var) SWIM_THROW_TYPE("an unspecified exception occurred, input type: ", var);
    #define SWIM_THROW_UNSPECIFIED SWIM_THROW_UNSPECIFIED_VAR(input);
    #define SWIM_THROW_EXISTING SWIM_RAISE_EXISTING
    """)



from __future__ import annotations

from typing import Callable, Type, Optional, Union, Pattern, Iterable, Any, TYPE_CHECKING, Set

from abc import ABC, abstractmethod
import re
from itertools import chain
from functools import partial

if TYPE_CHECKING:
    from swimport.model.cpp_object import Source, CPPObject

from swimport.__util__ import Replaceable

Swimporting = Callable[[Any], bool]


def object_swimporting(func: Callable[['Behaviour', 'SwimRule', Any, Any], Swimporting]) \
        -> Callable[['SwimRule', Any], Swimporting]:
    """
    a method wrapper that parts a function's last swim parameter, amking it create a delayed swimporting instead
    of running the function. The wrapped function also checks that the swim object hasn't imported the cpp object already.
    the wrapped function can also called normally as declared.
    """

    def ret(self, rule, obj, swim=...):
        def ret(swim):
            if obj in swim.imported:
                return False
            ret = func(self, rule, obj, swim)
            if ret or ret is None:
                swim.imported[obj] = rule
                return True
            return False

        if swim is not ...:
            return ret(swim)
        ret.object = obj
        ret.rule = rule
        return ret

    return ret


class Trigger(ABC, Replaceable):
    """An abstract class to decide whether a cpp object should be imported"""

    @abstractmethod
    def is_valid(self, rule, obj) -> bool:
        """get whether an object, under a specific rule, should be imported"""
        return True

    def __and__(self, other):
        """Create a trigger that only accepts values that both triggers accept"""
        return AndTrigger(self, other)

    def __or__(self, other):
        """Create a trigger that only accepts values that either triggers accept"""
        return OrTrigger(self, other)

    def __xor__(self, other):
        """Create a trigger that only accepts values that one of triggers accept, but not the other"""
        return XorTrigger(self, other)

    def __invert__(self):
        """Create a trigger that only accepts values that self does not"""
        return NotAnyTrigger(self)

    def __gt__(self, other):
        """Create a trigger that only accepts values that self accepts but other does not"""
        return GtTrigger(self, other)


class CategorizedTrigger(Trigger):
    """A custom trigger that only accept objects of a specific kind"""
    object_category: Type

    def is_valid(self, rule: 'SwimRule', obj):
        return super().is_valid(rule, obj) and isinstance(obj, self.object_category)

    def __init_subclass__(cls, *, specialization, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.object_category = specialization


Compileable = Union[str, Pattern[str]]


class NameTrigger(Trigger):
    """A trigger that only allows objects whose name fully matches the pattern"""

    def __init__(self, pattern: Compileable):
        self.name_pattern = re.compile(pattern)

    def is_valid(self, rule: 'SwimRule', obj):
        return super().is_valid(rule, obj) and self.name_pattern.fullmatch(obj.name)


class AggregateTrigger(Trigger):
    """A trigger that aggregates results of other sub-triggers"""

    def __init__(self, aggregate: Callable[[Iterable[bool]], bool], *triggers: Trigger):
        self.aggregate = aggregate
        self.sub_triggers = []
        if not triggers:
            raise ValueError('at least one trigger must be presented')
        object_category = ...
        for sub_trigger in triggers:
            if object_category:
                oc = getattr(sub_trigger, 'object_category', None)
                if object_category in (oc, ...):
                    object_category = oc
                else:
                    object_category = False

            # check if we can flatten another aggregate ((a|b)|(c|d)) -> (a|b|c|d)
            if isinstance(sub_trigger, AggregateTrigger) and sub_trigger.aggregate is self.aggregate:
                self.sub_triggers.extend(sub_trigger.sub_triggers)
            else:
                self.sub_triggers.append(sub_trigger)

        if object_category:
            self.object_category = object_category

    def is_valid(self, rule: 'SwimRule', obj):
        return self.aggregate(t.is_valid(rule, obj) for t in self.sub_triggers)


AndTrigger = partial(AggregateTrigger, all)
OrTrigger = partial(AggregateTrigger, any)
XorTrigger = partial(AggregateTrigger, lambda a: sum(map(bool, a)) % 2 == 1)
NotAnyTrigger = partial(AggregateTrigger, lambda a: not any(a))
GtTrigger = partial(AggregateTrigger, lambda i: bool(next(i)) > bool(next(i)))


class SwimRule(ABC):
    """A rule for swimporting"""

    @abstractmethod
    def do(self, obj) -> Optional[Swimporting]:
        """Get a swimporting for a cppobject, or None if the object does not pass the trigger"""
        pass

    def __call__(self, *sources: Source) -> Iterable[Swimporting]:
        """Get swimportings for all valid objects in sources"""
        if len(sources) > 1:
            src = chain.from_iterable(sources)
        elif sources:
            src = sources[0]
        else:
            return

        for o in src:
            s = self.do(o)
            if s:
                yield s


class Behaviour(SwimRule, Replaceable):
    """An abstract class to wrap objects to import in swimportings"""

    rule_cls: Type[TriggerBehaviourSwimRule]  # the type of rule to create when >> is used

    @abstractmethod
    def wrap(self, rule: SwimRule, obj) -> Swimporting:
        pass

    def default_is_valid(self, rule: 'SwimRule', obj) -> bool:
        return True

    def do(self, obj):
        if not self.default_is_valid(self, obj):
            return None
        return self.wrap(self, obj)

    def __rrshift__(self, other) -> 'TriggerBehaviourSwimRule':
        return self.rule_cls(other, self)


class CategorizedBehaviour(Behaviour):
    """A custom trigger to wrap objects of a specific kind. This class doesn't behave in any way special,
    but sets its inheritor's object_category so they can be used with categorised triggers"""
    object_category: Type[CPPObject]

    @abstractmethod
    def wrap(self, rule: 'SwimRule', obj) -> Swimporting:
        pass

    def default_is_valid(self, rule: 'SwimRule', obj):
        return super().default_is_valid(rule, obj) \
               and isinstance(obj, self.object_category)

    def __rrshift__(self, other):
        if (other is not ...) and (not isinstance(other, Trigger)):
            other = self.object_category.Trigger(other)
        return super().__rrshift__(other)

    def __init_subclass__(cls, *, specialization, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.object_category = specialization


class TriggerBehaviourSwimRule(SwimRule, Replaceable):
    """A rule for swimporting specific objects"""

    def __init__(self, trigger: Trigger, behaviour: Behaviour, name=None):
        """
        constructor
        :param trigger: the trigger of the rule, ... to accept any object
        :param behaviour: the behaviour to use
        :param name: the name of the rule (optional)
        """
        self.trigger = trigger
        self.behaviour = behaviour
        self.name = name
        t_category = getattr(self.trigger, 'object_category', None)
        b_category = getattr(self.behaviour, 'object_category', None)
        if t_category is not b_category \
                and self.trigger is not ...:
            raise ValueError(
                f'trigger and behaviour category mismatch: {t_category}(trigger) vs {b_category}(behaviour)')
        self.category = t_category

    def __str__(self):
        if self.category:
            if self.name:
                return f'{self.category.__name__} rule: {self.name}'
            return f'unnamed {self.category.__name__} rule'
        if self.name:
            return f'rule: {self.name}'
        return super().__str__()

    def do(self, obj) -> Optional[Swimporting]:
        """Get a swimporting for a cppobject, or None if the object does not pass the trigger"""
        if self.trigger is ...:
            valid = self.behaviour.default_is_valid(self, obj)
        else:
            valid = self.trigger.is_valid(self, obj)

        if valid:
            return self.behaviour.wrap(self, obj)
        return None


Behaviour.rule_cls = TriggerBehaviourSwimRule

class SwimportingsResult:
    """A result for importing multiple swimportings."""

    def __init__(self):
        self.value = 0
        self.names: Set[str] = set()

    def __iadd__(self, other):
        if isinstance(other, int):
            self.value += other
        elif isinstance(other, str):
            self.names.add(other)
            self.value += 1
        elif isinstance(other, Iterable):
            prev_len = len(self.names)
            self.names.update(other)
            self.value += (len(self.names) - prev_len)
        elif isinstance(other, SwimportingsResult):
            anon = other.value - len(other.names)
            prev_len = len(self.names)
            self.names.update(other.names)
            self.value += (len(self.names) - prev_len) + anon
        else:
            return NotImplemented
        return self

    def __bool__(self):
        return bool(self.value)

    def __eq__(self, other):
        return self.value == other \
               or (isinstance(other, type(self)) and other.value == self.value and other.names == self.names)

    def __ne__(self, other):
        return self.value != other \
               and (not isinstance(other, type(self)) or other.value != self.value or other.names != self.names)

    def __lt__(self, other):
        return self.value < other

    def __le__(self, other):
        return self.value <= other

    def __gt__(self, other):
        return self.value > other

    def __ge__(self, other):
        return self.value >= other

    def __contains__(self, item):
        return item in self.names

    def __int__(self):
        return self.value

    def __iter__(self):
        return (yield from self.names)


class SwimportAggregate(ABC):
    @abstractmethod
    def apply_porting(self, swimporting: Swimporting) -> Union[str, int]:
        """
        Apply a single porting to the swim
        :returns: whether or not the porting succeeded
        """

    def apply_portings(self, portings: Iterable[Swimporting]) -> SwimportingsResult:
        """
        apply multiple swimportings
        :return: a SwimportResult object
        """
        ret = SwimportingsResult()
        for p in portings:
            ret += self.apply_porting(p)
        return ret

    def __call__(self, arg) -> SwimportingsResult:
        """apply either a single or multiple swimportings"""
        if not isinstance(arg, Iterable):
            arg = arg,
        return self.apply_portings(arg)



import swimport.pools.derived_types.callable_
import swimport.pools.derived_types.iter_
import swimport.pools.derived_types.map_
import swimport.pools.derived_types.py_iterable
import swimport.pools.derived_types.slice_
import swimport.pools.derived_types.tuple_
import swimport.pools.derived_types.array



from swimport.pools.pools import pools
from swimport.pools.types import TypeSwimportingPool
from swimport.typeswim import TypeSwimporting, FunctionBody


@pools.add(TypeSwimportingPool)
def input_iterable(inner_type, *, cpp_name='py_iterable<?>', converter_name='py_iter::converter_to_cpp<?>',
                   deleter_name='py_iter::deleter<?>', swim, **type_args):
    """
    typemaps for input iterables
    NOTE: this pool can be applied automatically on most swimported types
    :param inner_type: the type to wrap in iterable input
    :param cpp_name: the name of the type to use in maps.
    :param converter_name: the name of the type of the converter function type.
    :param deleter_name: the name of the type of the deleter function type.
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    type_args.setdefault('in_iterable_types', ())
    type_args.setdefault('out_iterable_types', ())
    type_args.setdefault('inout_maps', False)

    inner_type, inner_porting = swim.get_porting(inner_type)
    if not inner_porting.to_cpp_func:
        raise Exception(f'type {inner_type} has no to_cpp function')

    cpp_name = cpp_name.replace('?', inner_type)
    converter_name = converter_name.replace('?', inner_type)
    deleter_name = deleter_name.replace('?', inner_type)

    frags = list(inner_porting.to_cpp_func.frags)

    if inner_porting.to_cpp_post_func:
        frags.extend(inner_porting.to_cpp_post_func.frags)

    return TypeSwimporting(cpp_name, py_name=f'Iterable[{inner_porting.py_name}]',
                           to_cpp=FunctionBody(f"""
                                                auto iter = PyObject_GetIter(input);
                                                if (!iter){{
                                                    SWIM_RAISE_TYPE("expected an iterable, not ", input);
                                                }}
                                                Py_DECREF(iter);
                                                return {{input, ({converter_name})PACKED_UD(0), ({deleter_name})PACKED_UD(1)}};
                                               """,
                                               user_data=f"""
                                                void* volatile ud[2];
                                                ud[0] = ({converter_name})([](PyObject* obj, int& ok)
                                                {{
                                                    void* volatile userdata=nullptr;
                                                    {inner_porting.to_cpp_func.user_data}
                                                    return {inner_porting.to_cpp_func.function_name}(obj, ok, userdata);
                                                }});
                                               """
                                                         + (f"""
                                                            ud[1] = ({deleter_name})([] ({inner_type} const & obj)
                                                            {{
                                                            int ok = 1;
                                                            void* volatile userdata=nullptr;
                                                            {inner_porting.to_cpp_post_func.user_data}
                                                            {inner_porting.to_cpp_post_func.function_name}(obj, ok, userdata);
                                                            }});
                                                            """ if inner_porting.to_cpp_post_func else
                                                            """
                                                            ud[1] = nullptr;
                                                            """)
                                                         + """
                                               userdata = (void * volatile)ud; 
                                               """,
                                               fragments=frags),
                           to_cpp_check="""
                                        auto iter = PyObject_GetIter(input);
                                        auto ret = (iter != nullptr);
                                        Py_XDECREF(iter);
                                        PyErr_Clear();
                                        return ret;
                                        """,
                           **type_args)



from swimport.pools.pools import pools
from swimport.pools.types import TypeSwimportingPool
from swimport.pools.derived_types.iter_ import iter_
from swimport.typeswim import TypeSwimporting, FunctionBody


@pools.add(TypeSwimportingPool, name='map')
def map_(key_type, value_type, *, cpp_name='std::unordered_map<?>', output=dict, swim, **type_args):
    """
    typemaps for using std::tuple for a set of types
    :param key_type: the key type to handle
    :param value_type: the value type to handle
    :param cpp_name: the name of the cpp type to map for. ? will be replaced with the type names
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """

    key_type, key_porting = swim.get_porting(key_type)
    value_type, value_porting = swim.get_porting(value_type)

    cpp_name = cpp_name.replace('?', key_type + ', ' + value_type)

    if key_porting.to_cpp_func and value_porting.to_cpp_func:
        to_cpp = FunctionBody.combine(key_porting.to_cpp_func, value_porting.to_cpp_func)(
            f"""
        void* kud = PACKED_UD(0);
        void* vud = PACKED_UD(1);
        {cpp_name} ret;

        auto keys = PyObject_CallMethod(input, "keys", nullptr);
        if (!keys){{
            PyErr_Clear();
            auto iter = PyObject_GetIter(input);
            if (!iter){{
                SWIM_RAISE_TYPE("expected an iterable, not ", input);
            }}
            PyObject* py_current;
            auto key_index = PyLong_FromLong(0);
            auto value_index = PyLong_FromLong(1);
            while (py_current = PyIter_Next(iter)){{
                auto key = PyObject_GetItem(py_current, key_index);
                if (!key) SWIM_RAISE_TYPE("expected 2-sequence in iterable, not ",py_current);
                auto value = PyObject_GetItem(py_current, value_index);
                if (!value) SWIM_RAISE_TYPE("expected 2-sequence in iterable, not ",py_current);
                auto cpp_k = {key_porting.to_cpp_func.function_name}(key, ok, kud);
                if (!ok || PyErr_Occurred()) SWIM_RAISE_UNSPECIFIED;
                ret[cpp_k] = {value_porting.to_cpp_func.function_name}(value, ok, vud);
                if (!ok || PyErr_Occurred()) SWIM_RAISE_UNSPECIFIED;
            }}
            Py_DECREF(iter);
            return ret;
        }}
        auto iter = PyObject_GetIter(keys);
        if (!iter){{
            SWIM_RAISE_TYPE("expected an iterable, not ", input);;
        }}
        PyObject* key;
        while (key = PyIter_Next(iter)){{
            auto value = PyObject_GetItem(input, key);
            if (!value) SWIM_RAISE_KEY("could not find key in mapping: ", key);
            auto cpp_k = {key_porting.to_cpp_func.function_name}(key, ok, kud);
            if (!ok || PyErr_Occurred()) SWIM_RAISE_UNSPECIFIED;
            ret[cpp_k] = {value_porting.to_cpp_func.function_name}(value, ok, vud);
            if (!ok || PyErr_Occurred()) SWIM_RAISE_UNSPECIFIED;
        }}
        Py_DECREF(iter);
        Py_DECREF(keys);
        return ret;
        """)

        to_cpp_check = """
        auto keys = PyObject_CallMethod(input, "keys", nullptr);
        PyObject* iter;
        if (keys){
            iter = PyObject_GetIter(keys);
        }
        else{
            iter = PyObject_GetIter(input);
        }
        bool ret = iter == nullptr;
        Py_XDECREF(keys);
        Py_XDECREF(iter);
        return ret;
        """
        if key_porting.to_cpp_post_func or value_porting.to_cpp_post_func:
            to_cpp_post = FunctionBody.combine(key_porting.to_cpp_post_func, value_porting.to_cpp_post_func)(
                f"""
            void* kud = PACKED_UD(0);
            void* vud = PACKED_UD(1);
            for (auto pair: input){{
            """
                + (
                    f'{key_porting.to_cpp_post_func.function_name}(pair.first, ok, kud); if (!ok || PyErr_Occurred()) return;'
                    if key_porting.to_cpp_post_func else '')
                + (
                    f'{value_porting.to_cpp_post_func.function_name}(pair.second, ok, vud); if (!ok || PyErr_Occurred()) return;'
                    if value_porting.to_cpp_post_func else '')
                + """
            }}
            """)
        else:
            to_cpp_post = None
    else:
        to_cpp = to_cpp_check = to_cpp_post = None

    if key_porting.to_py_func and value_porting.to_py_func:
        if output is dict:
            to_py = FunctionBody.combine(key_porting.to_py_func, value_porting.to_py_func)(f"""
                void* kud = PACKED_UD(0);
                void* vud = PACKED_UD(1);
                auto ret = PyDict_New();
                for (auto pair: input){{
                    auto key = {key_porting.to_py_func.function_name}(pair.first, ok, kud);
                    if (!ok || PyErr_Occurred()) {{ return ret; }}
                    auto value = {value_porting.to_py_func.function_name}(pair.second, ok, vud);
                    if (!ok || PyErr_Occurred()) {{ return ret; }}
                    if(PyDict_SetItem(ret ,key, value) != 0) {{ ok = false; return ret;}}
                }}
                return ret;
            """)
        elif output in (iter, list, tuple, set, frozenset):
            yield pools.tuple.get(key_type, value_type, swim=swim, pair=True)
            pair_type = 'std::pair<' + key_type + ', ' + value_type + '>'
            sub: TypeSwimporting = iter_.get(pair_type, cpp_name=cpp_name, input_=False, swim=swim, output=output)
            to_py = sub.to_py
        elif not output:
            to_py = None
        else:
            raise Exception('illegal output kind')
    else:
        to_py = None

    yield TypeSwimporting(cpp_name, 'Dict[' + key_porting.py_name + ', ' + value_porting.py_name + ']',
                          to_cpp=to_cpp,
                          to_py=to_py,
                          to_cpp_check=to_cpp_check,
                          to_cpp_post=to_cpp_post, **type_args)



from typing import Type, Union
from enum import Enum

from swimport.typedefswim import TypedefBehaviour
from swimport.typeswim import TypeSwimporting
from swimport.pools import syspools, pools
from swimport.model import NameTrigger, Enumeration, object_swimporting, Typedef


@Enumeration.set_default_trigger
class EnumerationTrigger(NameTrigger, Enumeration.Trigger):
    pass


@Enumeration.set_default_behaviour
class EnumerationBehaviour(Enumeration.Behaviour):
    default_typedef_behaviour = TypedefBehaviour(no_declare=True)

    def __init__(self, typedef=..., typedef_behaviour=...):
        """
        :param typedef: typedef to apply to the enum, if any. default is to extract from the enum object.
        """
        self.typedef = typedef
        if typedef_behaviour is ...:
            typedef_behaviour = self.default_typedef_behaviour
        self.typedef_behaviour = typedef_behaviour

    @object_swimporting
    def wrap(self, rule, obj: Enumeration, swim):
        swim(pools.primitive())
        swim.add_comment(obj)
        swim.add_raw(obj.body)
        if self.typedef:
            if self.typedef is ...:
                self.typedef = obj.type
            typedef_swimporting = self.typedef_behaviour.do(Typedef.make(obj.name, self.typedef))
            swim.apply_porting(typedef_swimporting)


class EnumBehaviour(EnumerationBehaviour):
    """A behaviour to wrap an enumeration in a python enum class"""
    # each value is a tuple of the form:
    # (Py_BuildValue format character, PyLong_As*)
    _typedef_bindings = {
        'char': ('b', 'PyLong_AsLong'),
        'unsigned char': ('B', 'PyLong_AsUnsignedLong'),
        'short int': ('h', 'PyLong_AsLong'),
        'unsigned short int': ('H', 'PyLong_AsUnsignedLong'),
        'int': ('i', 'PyLong_AsLong'),
        'unsigned int': ('I', 'PyLong_AsUnsignedLong'),
        'long int': ('l', 'PyLong_AsLong'),
        'unsigned long int': ('k', 'PyLong_AsUnsignedLong'),
        'long long int': ('L', 'PyLong_AsLongLong'),
        'unsigned long long int': ('K', 'PyLong_AsUnsignedLongLong'),
    }
    _default_bindings = _typedef_bindings['int']

    def __init__(self, super_cls: Union[str, Type[Enum]] = 'enum.IntEnum', del_from_ns=True):
        """
        :param super_cls: the super class of the python enum.
        :param del_from_ns: whether to delete the globals constants in the module namespace.
        """
        super().__init__(typedef=None)  # no typedefs, we're gonna make our own typemaps for the python enum
        self.super_cls = super_cls
        self.del_from_ns = del_from_ns

    def module_and_type_names(self):
        if isinstance(self.super_cls, type):
            return self.super_cls.__module__, self.super_cls.__name__
        sep_index = self.super_cls.find('.')
        if sep_index < 0:
            # default module is enum
            return 'enum', self.super_cls
        return self.super_cls[:sep_index], self.super_cls[sep_index + 1:]

    @object_swimporting
    def wrap(self, rule: 'SwimRule', obj, swim):
        super().wrap(rule, obj, swim)

        # the enum must be on the cxx side so that the python class object will be present at compile time,
        # allowing the typemaps to bind
        enum_id = swim.create_unique_id('py_enum_' + obj.name)
        build_value_char, py_to_c_func = self._typedef_bindings.get(obj.type, self._default_bindings)
        mem_decl_list = ', '.join('"' + str(mem.name) + '", ' + str(mem.name) for mem in obj.members)
        call_format_string = '"s(' + (f"(s{build_value_char})" * len(obj.members)) + ')"'
        sup_mod, sup_cls = self.module_and_type_names()
        swim.add_insert(f"""
            PyObject* {enum_id}_cls;
        """)
        swim.add_init(f"""
            auto {enum_id}_enum_module = PyImport_ImportModule("{sup_mod}");

            if (!{enum_id}_enum_module)
                return nullptr;
            
            {enum_id}_cls = PyObject_CallMethod({enum_id}_enum_module
                , "{sup_cls}"
                , {call_format_string}
                , "{obj.name}"
                , {mem_decl_list});

            if (!{enum_id}_cls)
                return nullptr;
            
            Py_DECREF({enum_id}_enum_module);
            SWIG_Python_SetConstant(d, "{enum_id}_cls",{enum_id}_cls);
        """)
        swim.add_python(f"""
            {obj.name} = _example.{enum_id}_cls;
        """)

        type_swimporting = TypeSwimporting(obj.name, obj.name,
                                           to_py=f"""
                                            return PyObject_CallFunction({enum_id}_cls, "({build_value_char})", input);
                                            """,
                                           to_cpp=f"""
                                            if (!PyObject_IsInstance(input, {enum_id}_cls))
                                            {{
                                                ok = false;
                                                PyErr_SetString(PyExc_TypeError,"expected {obj.name}");
                                                return {{}};
                                            }}
                                            auto v = PyObject_GetAttrString(input, "value");
                                            if (!v)
                                            {{
                                                ok = false;
                                                return {{}}; 
                                            }}
                                            auto ret = static_cast<{obj.name}>({py_to_c_func}(v));
                                            Py_DECREF(v);
                                            return ret;
                                            """,
                                           to_cpp_check=f"""
                                            return PyObject_IsInstance(input, {enum_id}_cls);
                                            """)
        swim(type_swimporting)

        if self.del_from_ns:
            swim(syspools.ns_trap(*(m.name for m in obj.members)))



from swimport.pools.pools import pools, syspools
from swimport.pools.types import TypeSwimportingPool
from swimport.typeswim import TypeSwimporting, FunctionBody


@syspools.add(TypeSwimportingPool)
def slice_no_step(start_type, end_type, cpp_name, *, swim, **type_args):
    start_type, start_porting = swim.get_porting(start_type)
    end_type, end_porting = swim.get_porting(end_type)

    cpp_name = cpp_name.replace('?', f'{start_type}, {end_type}')

    if start_porting.to_py_func and end_porting.to_py_func:
        body = f"""
        ok = 1;
        PyObject* start = input.has_start ? 
            {start_porting.to_py_func.function_name}(input.start, ok, PACKED_UD(0)) :
             nullptr;
        if (!ok || PyErr_Occurred() != nullptr)
            return nullptr;
        PyObject* end = input.has_start ? 
            {end_porting.to_py_func.function_name}(input.end, ok, PACKED_UD(1)) :
             nullptr;
        if (!ok || PyErr_Occurred() != nullptr){{
            Py_DECREF(start);
            return nullptr;
        }}
        PyObject* ret = PySlice_New(start, end, nullptr);
        Py_XDECREF(start);
        Py_XDECREF(end);
        return ret;
        """
        to_py = FunctionBody.combine(start_porting.to_py_func, end_porting.to_py_func)(body)
    else:
        to_py = None

    if start_porting.to_cpp_func and end_porting.to_cpp_func:
        body = f"""
        ok = 1;
        if (!PySlice_Check(input))
            SWIM_RAISE_TYPE("expected a slice, not ", input);
        PyObject* step = PyObject_GetAttrString(input, "step");
        if (step != Py_None)
        {{
            Py_DECREF(step);
            SWIM_RAISE(PyExc_ValueError, "slice cannot have a specified step");
        }}
        Py_DECREF(step);
        {cpp_name} ret;

        PyObject* start = PyObject_GetAttrString(input, "start");
        if (start == Py_None)
            ret.has_start = false;
        else{{
            ret.start = {start_porting.to_cpp_func.function_name}(start, ok, PACKED_UD(0));
            if (!ok || PyErr_Occurred() != nullptr) {{
                Py_DECREF(start);
                return ret;
            }}
            ret.has_start = true;
        }}
        Py_DECREF(start);

        PyObject* end = PyObject_GetAttrString(input, "stop");
        if (end == Py_None)
            ret.has_end = false;
        else{{
            ret.end = {end_porting.to_cpp_func.function_name}(end, ok, PACKED_UD(1));
            if (!ok || PyErr_Occurred() != nullptr) {{
                Py_DECREF(end);
                return ret;
            }}
            ret.has_end = true;
        }}
        Py_DECREF(end);

        return ret;
        """
        to_cpp = FunctionBody.combine(start_porting.to_cpp_func, end_porting.to_cpp_func)(body)
    else:
        to_cpp = None

    if start_porting.to_cpp_check_func and end_porting.to_cpp_check_func:
        body = f"""
        if (!PySlice_Check(input))
            return 0;
        PyObject* step = PyObject_GetAttrString(input, "step");
        if (step != Py_None)
        {{
            Py_DECREF(step);
            return 0;
        }}
        Py_DECREF(step);

        PyObject* start = PyObject_GetAttrString(input, "start");
        if (start != Py_None)
        {{
            if (!{start_porting.to_cpp_check_func.function_name}(start, PACKED_UD(0))){{
                Py_DECREF(start);
                return 0;
            }}
        }}
        Py_DECREF(start);

        PyObject* end = PyObject_GetAttrString(input, "stop");
        if (end != Py_None){{
            if (!{end_porting.to_cpp_check_func.function_name}(end, PACKED_UD(1))){{
                Py_DECREF(end);
                return 0;
            }}
        }}
        Py_DECREF(end);

        return 1;
        """
        to_cpp_check = FunctionBody.combine(start_porting.to_cpp_check_func, end_porting.to_cpp_check_func)(body)
    else:
        to_cpp_check = None

    if start_porting.to_cpp_post_func or end_porting.to_cpp_post_func:
        body = f"""
        ok = 1;
        """
        if start_porting.to_cpp_post_func:
            body += f"""
            if (input.has_start){{
                {start_porting.to_cpp_post_func}(input.start, ok, PACKED_UD(0));
            }}
            """
        if end_porting.to_cpp_post_func:
            body += f"""
            if (input.has_end){{
                {end_porting.to_cpp_post_func}(input.end, ok, PACKED_UD(1));
            }}
            """
        to_cpp_post = FunctionBody.combine(start_porting.to_cpp_post_func, end_porting.to_cpp_post_func)(body)
    else:
        to_cpp_post = None

    return TypeSwimporting(cpp_name, 'slice',
                           to_py=to_py,
                           to_cpp=to_cpp,
                           to_cpp_check=to_cpp_check,
                           to_cpp_post=to_cpp_post,
                           **type_args)


@syspools.add(TypeSwimportingPool)
def slice_step(start_type, end_type, step_type, cpp_name, *, swim, **type_args):
    start_type, start_porting = swim.get_porting(start_type)
    end_type, end_porting = swim.get_porting(end_type)
    step_type, step_porting = swim.get_porting(step_type)

    cpp_name = cpp_name.replace('?', f'{start_type}, {end_type}, {step_type}')

    if start_porting.to_py_func and end_porting.to_py_func and step_porting.to_py_func:
        body = f"""
        ok = 1;
        PyObject* start = input.has_start ? 
            {start_porting.to_py_func.function_name}(input.start, ok, PACKED_UD(0)) :
             nullptr;
        if (!ok || PyErr_Occurred() != nullptr)
            return nullptr;
        PyObject* end = input.has_start ? 
            {end_porting.to_py_func.function_name}(input.end, ok, PACKED_UD(1)) :
             nullptr;
        if (!ok || PyErr_Occurred() != nullptr){{
            Py_DECREF(start);
            return nullptr;
        }}
        PyObject* step = input.has_step ? 
            {step_porting.to_py_func.function_name}(input.step, ok, PACKED_UD(2)) :
             nullptr;
        if (!ok || PyErr_Occurred() != nullptr){{
            Py_DECREF(start);
            Py_DECREF(end);
            return nullptr;
        }}
        PyObject* ret = PySlice_New(start, end, step);
        Py_XDECREF(start);
        Py_XDECREF(end);
        Py_XDECREF(step);
        return ret;
        """
        to_py = FunctionBody.combine(start_porting.to_py_func, end_porting.to_py_func, step_porting.to_py_func)(body)
    else:
        to_py = None

    if start_porting.to_cpp_func and end_porting.to_cpp_func and step_porting.to_cpp_func:
        body = f"""
        ok = 1;
        if (!PySlice_Check(input))
            SWIM_RAISE_TYPE("expected a slice, not ", input);
        {cpp_name} ret;

        PyObject* start = PyObject_GetAttrString(input, "start");
        if (start == Py_None)
            ret.has_start = false;
        else{{
            ret.start = {start_porting.to_cpp_func.function_name}(start, ok, PACKED_UD(0));
            if (!ok || PyErr_Occurred() != nullptr) {{            
                Py_DECREF(start);
                return ret;
            }}
            ret.has_start = true;
        }}
        Py_DECREF(start);

        PyObject* end = PyObject_GetAttrString(input, "stop");
        if (end == Py_None)
            ret.has_end = false;
        else{{
            ret.end = {end_porting.to_cpp_func.function_name}(end, ok, PACKED_UD(1));
            if (!ok || PyErr_Occurred() != nullptr) {{
                Py_DECREF(end);
                return ret;
            }}
            ret.has_end = true;
        }}
        Py_DECREF(end);

        PyObject* step = PyObject_GetAttrString(input, "step");
        if (step == Py_None)
            ret.has_step = false;
        else{{
            ret.step = {step_porting.to_cpp_func.function_name}(step, ok, PACKED_UD(2));
            if (!ok || PyErr_Occurred() != nullptr) {{
                Py_DECREF(step);
                return ret;
            }}
            ret.has_step = true;
        }}
        Py_DECREF(step);

        return ret;
        """
        to_cpp = FunctionBody.combine(start_porting.to_cpp_func, end_porting.to_cpp_func, step_porting.to_cpp_func)(
            body)
    else:
        to_cpp = None

    if start_porting.to_cpp_check_func and end_porting.to_cpp_check_func:
        body = f"""
        if (!PySlice_Check(input))
            return 0;

        PyObject* start = PyObject_GetAttrString(input, "start");
        if (start != Py_None)
        {{
            if (!{start_porting.to_cpp_check_func.function_name}(start, PACKED_UD(0))){{
                Py_DECREF(start);
                return 0;
            }}
        }}
        Py_DECREF(start);

        PyObject* end = PyObject_GetAttrString(input, "stop");
        if (end != Py_None){{
            if (!{end_porting.to_cpp_check_func.function_name}(end, PACKED_UD(1))){{
                Py_DECREF(end);
                return 0;
            }}
        }}
        Py_DECREF(end);

        PyObject* step = PyObject_GetAttrString(input, "step");
        if (step != Py_None){{
            if (!{step_porting.to_cpp_check_func.function_name}(step, PACKED_UD(2))){{
                Py_DECREF(step);
                return 0;
            }}
        }}
        Py_DECREF(step);

        return 1;
        """
        to_cpp_check = FunctionBody.combine(start_porting.to_cpp_check_func, end_porting.to_cpp_check_func,
                                            step_porting.to_cpp_check_func)(body)
    else:
        to_cpp_check = None

    if start_porting.to_cpp_post_func or end_porting.to_cpp_post_func:
        body = f"""
        ok = 1;
        """
        if start_porting.to_cpp_post_func:
            body += f"""
            if (input.has_start){{
                {start_porting.to_cpp_post_func}(input.start, ok, PACKED_UD(0));
            }}
            """
        if end_porting.to_cpp_post_func:
            body += f"""
            if (input.has_end){{
                {end_porting.to_cpp_post_func}(input.end, ok, PACKED_UD(1));
            }}
            """
        if step_porting.to_cpp_post_func:
            body += f"""
            if (input.has_step){{
                {step_porting.to_cpp_post_func}(input.step, ok, PACKED_UD(2));
            }}
            """
        to_cpp_post = FunctionBody.combine(start_porting.to_cpp_post_func, end_porting.to_cpp_post_func,
                                           step_porting.to_cpp_post_func)(body)
    else:
        to_cpp_post = None

    return TypeSwimporting(cpp_name, 'slice',
                           to_py=to_py,
                           to_cpp=to_cpp,
                           to_cpp_check=to_cpp_check,
                           to_cpp_post=to_cpp_post,
                           **type_args)


@pools.add(TypeSwimportingPool, name='slice')
def slice_(start_type, end_type=..., step_type=None, cpp_name='slice<?>', *, swim, **type_args):
    """
    typemaps for using slices for a set of types
    :param start_type: the name of the imported type for the start of the slice
    :param end_type: the name of the imported type for the start of the slice. default is to use start_type
    :param end_type: the name of the imported type for the start of the slice. default is to not allow steps.
        ... to use start_type
    :param cpp_name: the name of the cpp type to map for. ? will be replaced with the type names
    :param type_args: keyword parameters forwarded to the TypeSwimporting created
    """
    if end_type is ...:
        end_type = start_type
    if step_type is ...:
        step_type = start_type

    if step_type:
        return slice_step(start_type, end_type, step_type, cpp_name, swim, **type_args)
    else:
        return slice_no_step(start_type, end_type, cpp_name, swim, **type_args)